

//////////////////////////////////////////////////////////////////
//                                                              //
//           PLINK (c) 2005-2008 Shaun Purcell                  //
//                                                              //
// This file is distributed under the GNU General Public        //
// License, Version 2.  Please see the file COPYING for more    //
// details                                                      //
//                                                              //
//////////////////////////////////////////////////////////////////


#include <iostream>
#include <fstream>
#include <iomanip>
#include <cmath>
#include <algorithm>
#include <map>

#include "plink.h"
#include "fisher.h"
#include "stats.h"

#include "../br2.h"
/*
 *
#include "helper.h"
#include "options.h"
#include "crandom.h"
#include "sets.h"
#include "perm.h"
*/


using namespace std;



/////////////////////////////////////////////
// Full model association tests

void Plink::fullModelAssoc(
		bool print_results,
		vector<Individual*> & sample ,
		 vector<Locus*> &locus,
		 vector<ModelOutput> &out
)
{
	out.clear();

  if (print_results)
    printLOG("Full-model association tests, minimum genotype count: --cell " +
	     int2str(par::min_geno_cell) + "\n");


  ofstream ASC;

  if (print_results)
    {
      string f = par::output_file_name + ".model";
      ASC.open(f.c_str(),ios::out);
      printLOG("Writing full model association results to [ " + f + " ] \n");

      ASC << setw(4) << "CHR" << " "
	  << setw(par::pp_maxsnp) << "SNP" << " "
	  << setw(4) << "A1" << " "
	  << setw(4) << "A2" << " "
	  << setw(8) << "TEST" << " "
	  << setw(14) << "AFF" << " "
	  << setw(14) << "UNAFF" << " ";
      if ( ! par::fisher_test )
	ASC << setw(12) << "CHISQ" << " "
	    << setw(4) << "DF" << " ";
      ASC << setw(12) << "P" << "\n";

      ASC.precision(4);
  }


#if 0
  vector<Individual*> cases, controls;
  for(int i = 0; i < sample.size() ; i ++)
      if(sample[i]->phenotype == 2) cases.push_back(sample[i]);
      else controls.push_back(sample[i]);
#endif

  ///////////////////////////////
  // Iterate over SNPs


  vector<Locus*>::iterator s = locus.begin();
  int l=0;

  out.clear();

  while ( s != locus.end() )
    {

      int A11=0, A12=0, A22=0;
      int U11=0, U12=0, U22=0;


      ////////////////////////
      // Autosomal or haploid?

      bool X=false, haploid=false;
    //  if (par::chr_sex[locus[l]->chr]) X=true;
    //  else if (par::chr_haploid[locus[l]->chr]) haploid=true;

      // Skip haploid markers
      if (haploid)
	{
	  s++;
	  l++;
	  continue;
	}


      /////////////////////////////
      // Iterate over individuals


     // vector<bool>::iterator i1 = (*s)->one.begin();
     // vector<bool>::iterator i2 = (*s)->two.begin();

      int l_id = locus[l]->id;

#if 0
      vector<Individual*>::iterator gperson = sample.begin();

      int A[4] = {0,0,0,0}, U[4] = {0,0,0,0};

      gperson = cases.begin();
      while ( gperson != cases.end() ) {
	  Individual * pperson = *gperson;

	  int geno = pperson->getGeno(l_id);

	  A[geno] ++;

	  gperson++;
      }

      gperson = controls.begin();
      while ( gperson != controls.end() ) {
	  Individual * pperson = *gperson;

	  int geno = pperson->getGeno(l_id);
	  U[geno] ++;

	  gperson++;
      }


      A11 = A[G_aa];
      A12 = A[G_aA];
      A22 = A[G_AA];
      U11 = U[G_aa];
      U12 = U[G_aA];
      U22 = U[G_AA];

#else

      vector<Individual*>::iterator gperson = sample.begin();


      while ( gperson != sample.end() )
	{

	  // Phenotype for this person (i.e. might be permuted)
	  //Individual * pperson = (*gperson)->pperson;
	  Individual * pperson = *gperson;


	  // SNP alleles

	  //bool s1 = pperson->one[l];
	  //bool s2 = pperson->two[l];
	  int xxx = pperson->getGeno(l_id);
	  bool s1 = xxx & 1;
	  bool s2 = xxx >> 1;



	  //if ( ! pperson->missing  )
	    {

	      // Only consider diploid chromosomes
	     // if ( ! ( X && (*gperson)->sex ) )
		{

		  if ( pperson->phenotype == 2 )     // cases
		    {
		      if ( ! s1 )
			{
			  if ( ! s2 )     // Homozyg 00
			    A11++;
			  else            // Hetero  01
			    A12++;
			}
		      else if ( s2 )      // Homozyg 11
			A22++;
		    }
		  else // if ( pperson->phenotype == 1 ) // controls
		    {
		      if ( !s1 )
			{
			  if ( !s2 )      // Homozyg 00
			    U11++;
			  else            // Hetero  01
			    U12++;
			}
		      else if ( s2 )      // Homozyg 11
			U22++;
		    }


		}
	    }


	  // Next individual
	  gperson++;
	//  i1++;
	//  i2++;

	}
#endif

      ///////////////////////////////////
      // Calculate association statistics

      double obs_A = A11 + A12 + A22;
      double obs_U = U11 + U12 + U22;
      double obs_T = obs_A + obs_U;

      double obs_1 = 2*(A11+U11) + A12 + U12;
      double obs_2 = 2*(A22+U22) + A12 + U12;

      double obs_11 = A11+U11;
      double obs_12 = A12+U12;
      double obs_22 = A22+U22;

      bool invalid = false;


      double gen_p, dom_p, rec_p;
      double dom_chisq, rec_chisq, gen_chisq;

      double mult_p, mult_chisq;
      double CA_chisq;
      double CA_p;


      if (A11 < par::min_geno_cell ||
	  A12 < par::min_geno_cell ||
	  A22 < par::min_geno_cell) invalid = true;
      else if (U11 < par::min_geno_cell ||
	       U12 < par::min_geno_cell ||
	       U22 < par::min_geno_cell) invalid = true;


      /* Now we have the contigency table */

      #if 0

      cout << "Table for snp " << (*s)->name << " " << "a1=" << (*s)->allele1 << "\n";
      cout << "  A: " << A11 << " " << A12 << " " << A22 << "\n";
      cout << "  U: " << U11 << " " << U12 << " " << U22 << "\n";
      cout << " Invalid = " << invalid << "\n";

      {
      double g0 = log((double)A11/U11);
      double g1 = log((double)A12*U11/((double)A11*U12));
      double g2 = log((double)A22*U11/((double)A11*U22));


      cout << "  g0=" << log((double)A11/U11) << "\n";
      cout << "  g1=" << log((double)A12*U11/((double)A11*U12)) << "\n";
      cout << "  g2=" << log((double)A22*U11/((double)A11*U22)) << "\n";

      cout << "ADD " << exp(0.5*g2) << "\n";
      cout << "DOMDEV " << exp(g1-0.5*g2) << "\n";

      }

      cout << "  alt g0=" << log((double)A22/U22) << "\n";
      cout << "  alt g1=" << log((double)A12*U22/((double)A22*U12)) << "\n";
      cout << "  alt g2=" << log((double)A11*U22/((double)A22*U11)) << "\n";

      double g0 = log((double)A22/U22);
      double g1 = log((double)A12*U22/((double)A22*U12));
      double g2 = log((double)A11*U22/((double)A22*U11));

      cout << "ADD " << exp(0.5*g2) << "\n";
      cout << "DOMDEV " << exp(g1-0.5*g2) << "\n";



      #endif




      ///////////////////////
      // Cochram-Armitage Trend test

      double CA = ( ( obs_U / obs_T * A12 ) - ( obs_A / obs_T * U12 ) )
	+ 2*( ( obs_U / obs_T * A22 ) - ( obs_A / obs_T * U22 ) ) ;

      double varCA = obs_A * obs_U
	* ( ( obs_T * ( obs_12 + 4*obs_22 )
	      - ( obs_12+2*obs_22 ) * ( obs_12+2*obs_22 )  )
	    / (obs_T * obs_T * obs_T  )) ;

      // #pragma omp critical
      {


       // This is the trend test
       CA_chisq = (CA*CA) / varCA;
       CA_p = chiprobP(CA_chisq,1);

      ///////////////////////
      // Multiplicative model

      double obs_A1 = 2*A11 + A12;
      double obs_A2 = 2*A22 + A12;
      double obs_U1 = 2*U11 + U12;
      double obs_U2 = 2*U22 + U12;

      if ( par::fisher_test )
	{
	  table_t t;
	  sizeTable(t,2,2);
	  t[0][0] = (int)obs_A1;
	  t[1][0] = (int)obs_A2;
	  t[0][1] = (int)obs_U1;
	  t[1][1] = (int)obs_U2;
	  mult_p = fisher(t);

	}
      else
	{

	  double exp_A1 = (obs_A * obs_1 ) / obs_T;    // note 2's cancelled for obs_A and obs_T
	  double exp_A2 = (obs_A * obs_2 ) / obs_T;    // which are counts of individuals, not
	  double exp_U1 = (obs_U * obs_1 ) / obs_T;    // alleles
	  double exp_U2 = (obs_U * obs_2 ) / obs_T;

	  mult_chisq =  ( ( obs_A1 - exp_A1 ) * ( obs_A1 - exp_A1 ) ) / exp_A1
	    + ( ( obs_A2 - exp_A2 ) * ( obs_A2 - exp_A2 ) ) / exp_A2
	    + ( ( obs_U1 - exp_U1 ) * ( obs_U1 - exp_U1 ) ) / exp_U1
	    + ( ( obs_U2 - exp_U2 ) * ( obs_U2 - exp_U2 ) ) / exp_U2;


	  ///////////////////////
	  // Multiplicative model

	  mult_p = chiprobP(mult_chisq,1);

	}


       gen_p = dom_p = rec_p = -9;


      if (!invalid)
	{

	  //////////////////////////////////////////////////////////////
	  // Standard chi-square test, or Fisher's exact

	  if ( par::fisher_test )
	    {

	      ////////////
	      // General

	      table_t t;
	      sizeTable(t,3,2);
	      t[0][0] = A11;
	      t[1][0] = A12;
	      t[2][0] = A22;
	      t[0][1] = U11;
	      t[1][1] = U12;
	      t[2][1] = U22;
	      gen_p = fisher(t);


	      ////////////
	      // Dominant

	      sizeTable(t,2,2);
	      t[0][0] = A11+A12;
	      t[1][0] = A22;
	      t[0][1] = U11+U12;
	      t[1][1] = U22;
	      dom_p = fisher(t);

	      /////////////
	      // Recessive

	      sizeTable(t,2,2);
	      t[0][0] = A11;
	      t[1][0] = A12+A22;
	      t[0][1] = U11;
	      t[1][1] = U12+U22;
	      rec_p = fisher(t);


	    }
	  else
	    {

	      ///////////////////////
	      // General model

	      double exp_A11 = (obs_A * obs_11 ) / obs_T;
	      double exp_A12 = (obs_A * obs_12 ) / obs_T;
	      double exp_A22 = (obs_A * obs_22 ) / obs_T;
	      double exp_U11 = (obs_U * obs_11 ) / obs_T;
	      double exp_U12 = (obs_U * obs_12 ) / obs_T;
	      double exp_U22 = (obs_U * obs_22 ) / obs_T;

	      gen_chisq =  ( ( A11 - exp_A11 ) * ( A11 - exp_A11 ) ) / exp_A11
		+ ( ( A12 - exp_A12 ) * ( A12 - exp_A12 ) ) / exp_A12
		+ ( ( A22 - exp_A22 ) * ( A22 - exp_A22 ) ) / exp_A22
		+ ( ( U11 - exp_U11 ) * ( U11 - exp_U11 ) ) / exp_U11
		+ ( ( U12 - exp_U12 ) * ( U12 - exp_U12 ) ) / exp_U12
		+ ( ( U22 - exp_U22 ) * ( U22 - exp_U22 ) ) / exp_U22;


	      ///////////////////////
	      // Dominant (minor allele) (1) model

	      dom_chisq =  ( ( (A11+A12) - (exp_A11+exp_A12) ) * ( (A11+A12) - (exp_A11+exp_A12) ) ) / (exp_A11+exp_A12)
		+ ( ( A22 - exp_A22 ) * ( A22 - exp_A22 ) ) / exp_A22
		+ ( ( (U11+U12) - (exp_U11+exp_U12) ) * ( (U11+U12) - (exp_U11+exp_U12) ) ) / (exp_U11+exp_U12)
		+ ( ( U22 - exp_U22 ) * ( U22 - exp_U22 ) ) / exp_U22;


	      //////////////////////////////////////
	      // Recessive (minor allele) (1) model

	      rec_chisq =  ( ( (A22+A12) - (exp_A22+exp_A12) ) * ( (A22+A12) - (exp_A22+exp_A12) ) ) / (exp_A22+exp_A12)
		+ ( ( A11 - exp_A11 ) * ( A11 - exp_A11 ) ) / exp_A11
		+ ( ( (U22+U12) - (exp_U22+exp_U12) ) * ( (U22+U12) - (exp_U22+exp_U12) ) ) / (exp_U22+exp_U12)
		+ ( ( U11 - exp_U11 ) * ( U11 - exp_U11 ) ) / exp_U11;


	      //////////////////////////////////
	      // p-values and model comparisons

	      gen_p = chiprobP(gen_chisq,2);
	      dom_p = chiprobP(dom_chisq,1);
	      rec_p = chiprobP(rec_chisq,1);

	    }

	}

      }



      if (print_results)
	{
	  ostream &ASC = cout;
	  /////////////
	  // Genotypic

	  ASC << setw(4) << locus[l]->chr << " "
	      << setw(par::pp_maxsnp) << locus[l]->name << " "
	      << setw(4) << locus[l]->allele1 << " "
	      << setw(4) << locus[l]->allele2 << " "
	      << setw(8) << "GENO" << " "
	      << setw(14) << int2str(A11)+"/"+int2str(A12)+"/"+int2str(A22)  << " "
	      << setw(14) << int2str(U11)+"/"+int2str(U12)+"/"+int2str(U22) << " " ;
	  if (gen_p < -1)
	    {
	      if ( ! par::fisher_test )
		ASC << setw(12) << "NA"  << " "
		    << setw(4)  << "NA" << " ";

	      ASC << setw(12) << "NA" << "\n" ;
	    }
	  else
	    {
	      if ( ! par::fisher_test )
		ASC << setw(12) << gen_chisq << " "
		    << setw(4)  << "2" << " ";

	      ASC << setw(12) << gen_p << "\n";
	    }

	  /////////////////
	  // CA trend test

	  ASC << setw(4) << locus[l]->chr << " "
	      << setw(par::pp_maxsnp) << locus[l]->name << " "
	      << setw(4) << locus[l]->allele1 << " "
	      << setw(4) << locus[l]->allele2 << " "
	      << setw(8) << "TREND" << " "
	      << setw(14) << int2str(A11*2+A12)+"/"+int2str(A12+A22*2) << " "
	      << setw(14) << int2str(U11*2+U12)+"/"+int2str(U12+U22*2) << " ";
	  if (CA_p < -1)
	    {
	      if ( ! par::fisher_test )
		ASC << setw(12) << "NA" << " "
		    << setw(4) << "NA" << " ";
	      ASC << setw(12) << "NA" << "\n" ;
	    }
	  else
	    {
	      if ( ! par::fisher_test )
		ASC << setw(12) << CA_chisq << " "
		    << setw(4) << "1" << " ";
	      ASC << setw(12) << CA_p << "\n" ;
	    }

	  /////////////
	  // Allelic

	  ASC << setw(4) << locus[l]->chr << " "
	      << setw(par::pp_maxsnp) << locus[l]->name << " "
	      << setw(4) << locus[l]->allele1 << " "
	      << setw(4) << locus[l]->allele2 << " "
	      << setw(8) << "ALLELIC" << " "
	      << setw(14) << int2str(A11*2+A12)+"/"+int2str(A12+A22*2) << " "
	      << setw(14) << int2str(U11*2+U12)+"/"+int2str(U12+U22*2) << " ";
	  if (mult_p < -1)
	    {
	      if ( ! par::fisher_test )
		ASC << setw(12) << "NA" << " "
		    << setw(4) << "NA" << " ";
	      ASC << setw(12) << "NA" << "\n" ;
	    }
	  else
	    {
	      if ( ! par::fisher_test )
		ASC << setw(12) << mult_chisq << " "
		    << setw(4) << "1" << " ";
	      ASC << setw(12) << mult_p << "\n" ;
	    }

	  /////////////
	  // Dominant

	  ASC << setw(4) << locus[l]->chr << " "
	      << setw(par::pp_maxsnp) << locus[l]->name << " "
	      << setw(4) << locus[l]->allele1 << " "
	      << setw(4) << locus[l]->allele2 << " "
	      << setw(8) << "DOM" << " "
	      << setw(14) << int2str(A11+A12)+"/"+int2str(A22) << " "
	      << setw(14) << int2str(U11+U12)+"/"+int2str(U22) << " ";
	  if (dom_p < -1)
	    {
	      if ( ! par::fisher_test )
		ASC << setw(12) << "NA" << " "
		    << setw(4) << "NA" << " ";
	      ASC << setw(12) << "NA" << "\n" ;
	    }
	  else
	    {
	      if ( ! par::fisher_test )
		ASC << setw(12) << dom_chisq << " "
		    << setw(4) << "1" << " ";
	      ASC << setw(12) << dom_p << "\n" ;
	    }

	  /////////////
	  // Recessive

	  ASC << setw(4) << locus[l]->chr << " "
	      << setw(par::pp_maxsnp) << locus[l]->name << " "
	      << setw(4) << locus[l]->allele1 << " "
	      << setw(4) << locus[l]->allele2 << " "
	      << setw(8) << "REC" << " "
	      << setw(14) << int2str(A11)+"/"+int2str(A12+A22) << " "
	      << setw(14) << int2str(U11)+"/"+int2str(U12+U22) << " ";
	  if (rec_p < -1)
	    {
	      if ( ! par::fisher_test )
		ASC << setw(12) << "NA" << " "
		    << setw(4) << "NA" << " ";
	      ASC << setw(12) << "NA" << "\n" ;
	    }
	  else
	    {
	      if ( ! par::fisher_test )
		ASC << setw(12) << rec_chisq << " "
		    << setw(4) << "1" << " ";
	      ASC << setw(12) << rec_p << "\n" ;
	    }
	}

      ModelOutput mod;

      mod.snp = locus[l];
      mod.a1 = locus[l]->allele1;
      mod.pGeno = gen_p < -1 ? NA : gen_p ;
      mod.pTrend = CA_p < -1 ? NA : CA_p ;

      if(invalid) mod.pGeno = NA;

      if(0) if(invalid) {
	  cout << "Warning: Sparse contigency table for SNP: " << locus[l]->name << " ";
	  cout << A11 << "/" << A12 << "/" << A22 << " ";
	  cout << U11 << "/" << U12 << "/" << U22 << " ";
	  cout << "\n";

      }
      out.push_back(mod);

     // cout << mod.snp->name << " " << mod.pGeno << " " << mod.pTrend << "\n";

      // Next SNP
      s++;
      l++;

    }

  if (print_results)
    {
      ASC.close();
    }

}




/////////////////////////////////////////////
// Simple quantitative trait association test
// note: does not explicitly treat X/haploid
// markers differently: see --linear

void Plink::testQAssoc(bool print_results,
    vector<Individual*> &Sample, vector<Locus*> &SNPs,
    vector <LogisticOutput> &results)

{
  int persons = 0;
  int locus = 2;

  ofstream ASC, QT_MEANS;
  results.clear();

  if (print_results)
    {
      string f = par::output_file_name + ".qassoc";
      printLOG("Writing QT association results to [ " + f + " ] \n");
      ASC.open(f.c_str(),ios::out);
      ASC << setw(4) << "CHR" << " "
	  << setw(par::pp_maxsnp) << "SNP" << " "
	  << setw(10) << "BP" << " "
	  << setw(8) << "NMISS" << " "
	  << setw(10) << "BETA" << " "
	  << setw(10) << "SE" << " "
	  << setw(10) << "R2" << " "
	  << setw(8) << "T" << " "
	  << setw(12) << "P" << " "
	  << "\n";
      ASC.precision(4);

    }


  ////////////////////////////
  // Iterate over each locus

  vector<Locus*>::iterator s = SNPs.begin();
  int l=0;

  while ( s != SNPs.end() )
    {

      double g_mean=0, g_var=0;
      double qt_mean=0, qt_var=0;
      double qt_g_covar=0;

      // number of individuals in analysis
      int nanal = 0;

      /////////////////////////////
      // Iterate over individuals

      //vector<bool>::iterator i1 = (*s)->one.begin();
      //vector<bool>::iterator i2 = (*s)->two.begin();
      vector<Individual*>::iterator gperson = Sample.begin();

      int l_id = SNPs[l]->id;

      while ( gperson != Sample.end() )
	{

	  // Phenotype for this person (i.e. might be permuted)

	  Individual * pperson = *gperson;

	  // SNP alleles
	  int a = pperson->getGeno(l_id);
	  // a = (one) (two)

	  // if(SNPs[l]->id != l_id) cout << "problem with qassoc: SNPs[" << l << "].id=" << SNPs[l]->id << " != " << l << "\n";

	  bool s2 = a >> 1;
	  bool s1 = a & 1;

	    {
	      if ( ! ( s1 && (!s2) ) )     //   10 = missing
		{
		  qt_mean += pperson->phenotype;

		  if (!s1)
		    {
		      if (!s2)      //   00 = hom(11)
			g_mean+=2;
		      else          //   01 = het(12)
			g_mean++;
		    }
		  nanal++;
		}

	    }


	  // Advance to the next person (for phenotype information
	  // and the two SNP alleles also)

	  gperson++;

	}

      qt_mean /= (double)nanal;
      g_mean /= (double)nanal;


      //////////////////////////////////
      // Iterate over individuals again

      //i1 = (*s)->one.begin();
      //i2 = (*s)->two.begin();
      gperson = Sample.begin();

      while ( gperson != Sample.end() )
	{

	  // Phenotype for this person (i.e. might be permuted)
	  Individual * pperson = *gperson;

          // SNP alleles
          int a = pperson->getGeno(l_id);
          // a = (one) (two)

          bool s2 = a >> 1;
          bool s1 = a & 1;

	    {
	      if ( ! ( (s1) && (!s2) ) )
		{

		  qt_var += (pperson->phenotype-qt_mean)
		    * ( pperson->phenotype-qt_mean ) ;

		  double g = 0;

		  if (!s1)
		    {
		      if (!s2)
			g=2;
		      else
			g=1;
		    }

		  g_var += (g-g_mean) * ( g-g_mean ) ;
		  qt_g_covar += ( pperson->phenotype - qt_mean )
		    * ( g - g_mean ) ;

		}
	    }

	  // Advance to the next person
	  gperson++;

	}

      // Summary statistics

      qt_var /= (double)nanal - 1;
      g_var /= (double)nanal - 1;
      qt_g_covar /= (double)nanal - 1;

      // Test statistics
      double beta = qt_g_covar / g_var;
      double vbeta = ( qt_var/g_var - (qt_g_covar*qt_g_covar)/(g_var*g_var) ) / (nanal-2);
      double t = beta / sqrt(vbeta);
      double t_p;

      // #pragma omp critical
      {

	  // Display results?
	  t_p = pT(t,nanal-2);

      }

      if(t_p >= 0 && realnum(beta))
        results.push_back(LogisticOutput(SNPs[l], SNPs[l]->allele1, pow(M_E,beta), t_p));
      else
        results.push_back(LogisticOutput(SNPs[l], SNPs[l]->allele1, NA, NA));


      if (print_results)
	{
         double r2 =   (qt_g_covar * qt_g_covar ) / ( qt_var * g_var ) ;

// 	  double wald = (beta*beta) / ( vbeta ) ;
// 	  double wald_p = chiprobP(wald,1);
//        double lrt = -nanal * log(1-r2);

	  ASC << setw(4) << SNPs[l]->chr  << " "
	      << setw(par::pp_maxsnp) << SNPs[l]->name << " "
	      << setw(10) << SNPs[l]->bp << " "
	      << setw(8) << nanal << " ";
	  if ( ! realnum(beta) )
	    {
	      ASC << setw(10) << "NA" << " "
		  << setw(10) << "NA" << " "
		  << setw(10) << "NA" << " " ;
	    }
	  else
	    {
	      ASC << setw(10) << beta << " "
		  << setw(10) << sqrt(vbeta) << " "
		  << setw(10) << r2 << " " ;
	    }

 	  if (t_p >= 0)
 	    ASC << setw(8) << t  << " " << setw(12) << t_p << " " ;
 	  else
 	    ASC << setw(8) << "NA"  << " " << setw(12) << "NA"  << " " ;

	  ASC << "\n";

	}


    skip_p2:

      // Next SNP
      s++;
      l++;
    }

  if (print_results)
    {
      ASC.close();
     }

}


/////////////////////////////////////////////////
// Basic C/C association test



void Plink::testAssoc(vector<Individual*> &sample, vector<Locus*> &locus,
		vector<LogisticOutput> &out)
{
    out.clear();
#ifdef DEBUG
   ofstream ASC;

    if(0) {
      ////////////////////
      // Display results

      string f = par::output_file_name + ".assoc";
      if ( par::fisher_test )
	f = par::output_file_name + ".fisher";
      ASC.open(f.c_str(),ios::out);
      ASC.precision(4);
      printLOG("Writing main association results to [ " + f + " ] \n");

      ASC << setw(4) << "CHR" << " "
	  << setw(par::pp_maxsnp)<< "SNP" << " "
	  << setw(10) << "BP" << " "
	  << setw(4) << "A1" << " "
	  << setw(8) << "F_A" << " "
	  << setw(8) << "F_U" << " "
	  << setw(4) << "A2" << " ";
      if ( ! par::fisher_test )
	ASC << setw(12)<< "CHISQ" << " ";
      ASC << setw(12)<< "P" << " "
	  << setw(12)<< "OR" << " " ;
      if (par::display_ci)
	ASC << setw(12) << string("L"+dbl2str(0.95*100)) << " "
	    << setw(12) << string("U"+dbl2str(0.95*100)) << " ";

      ASC << "\n";

    }
#endif


  ///////////////////////////////
  // Iterate over SNPs

  vector<Locus*>::iterator s = locus.begin();
  int l=0;

  while ( s != locus.end() )
    {
      int l_id = locus[l]->id;

      // In adaptive mode, possibly skip this test
      if (0)
	{
	  s++;
	  l++;
	  continue;
	}

      int A1 = 0, A2 = 0, A0 = 0;
      int U1 = 0, U2 = 0, U0 = 0;

      /////////////////////////////
      // Iterate over individuals

      vector<Individual*>::iterator gperson = sample.begin();

      while ( gperson != sample.end() )
	{
	  Individual * pperson = *gperson;

	  int xxx = pperson->getGeno(l_id);
	  bool s1 = xxx & 1;
	  bool s2 = xxx >> 1;

	  // Is this individual missing?
	  // WARN: Assume not missing here.

	      /////////////////////////////////////
	      // Autosomal marker
	      if ( pperson->phenotype == 2 ) // Affected
		{
		  if (!s1)
		    {
		      if (!s2)
			A1+=2;
		      else
			{ A1++; A2++; }
		    }
		  else
		    {
		      if ( s2 )
			A2+=2;
		      else A0+=2;
		    }
		}
	      else  // unaffected if not missing
		{
		  if (!s1)
		    {
		      if (!s2)
			U1+=2;
		      else
			{ U1++; U2++; }
		    }
		  else
		    {
		      if ( s2 )
			U2+=2;
		      else U0+=2;
		    }
		}


	  // Next person
	  gperson++;
	}


      // Calculate standard association statistic

      // Total number of alleles
      double tot = A1+A2+U1+U2;

      double pvalue;
      double odds, original, exp_afffreq1, exp_afffreq2, exp_unffreq1, exp_unffreq2;
      int a0, a1, a2;

      // Total number of non-missing affecteds/unaffecteds
      if ( par::fisher_test )
	{
	  table_t t;
	  sizeTable(t,2,2);
	  t[0][0] = A1;
	  t[0][1] = U1;
	  t[1][0] = A2;
	  t[1][1] = U2;
	  pvalue = fisher(t);
	  original = 1 - pvalue;
	}
      else if (1 /*!par::assoc_test_alt_perm*/)
	{
	  double Taff = A1+A2;
	  double Tunf = U1+U2;

	  double Ta1 = A1+U1;
	  double Ta2 = A2+U2;

	  double Texp_afffreq1 = (Taff*Ta1)/tot;
	  double Texp_unffreq1 = (Tunf*Ta1)/tot;
	  double Texp_afffreq2 = (Taff*Ta2)/tot;
	  double Texp_unffreq2 = (Tunf*Ta2)/tot;

 	  original = ((A1 - Texp_afffreq1) * (A1 - Texp_afffreq1)) / Texp_afffreq1
 	    + ( (U1 - Texp_unffreq1) * ( U1 - Texp_unffreq1 ) ) / Texp_unffreq1
 	    + ( (A2 - Texp_afffreq2) * ( A2 - Texp_afffreq2 ) ) / Texp_afffreq2
 	    + ( (U2 - Texp_unffreq2) * ( U2 - Texp_unffreq2 ) ) / Texp_unffreq2 ;

	}
      else // if (par::assoc_test_alt_perm)
	{

	  // Total number of non-missing affecteds/unaffecteds
	  int aff = A1+A2;
	  int unf = U1+U2;

	  a1 = A1+U1;
	  a2 = A2+U2;
	  a0 = A0+U0;

	  exp_afffreq1 = (aff*a1)/tot;
	  exp_unffreq1 = (unf*a1)/tot;
	  exp_afffreq2 = (aff*a2)/tot;
	  exp_unffreq2 = (unf*a2)/tot;

	  // Include missing alleles for final marginal values
	  aff += A0;
	  unf += U0;

	  original = ((A1 - exp_afffreq1) * (A1 - exp_afffreq1)) / exp_afffreq1
	    + ( (U1 - exp_unffreq1) * ( U1 - exp_unffreq1 ) ) / exp_unffreq1
	    + ( (A2 - exp_afffreq2) * ( A2 - exp_afffreq2 ) ) / exp_afffreq2
	    + ( (U2 - exp_unffreq2) * ( U2 - exp_unffreq2 ) ) / exp_unffreq2 ;

	}


      ////////////////////////////////////////////////////
      // Do we need to calculate odds ratio and p-value?
      // (Either for display (of original data) or for
      // other reasons (profile set-based test)

      if ( 1 )
	{
	  // Note: in set-score mode, use an adjusted form
	  // of the odds-ratio, to ensure it is always valud

	  if ( 0 /* par::set_score*/ )
	    odds = (double)( (A1+0.5)*(U2+0.5) ) / (double)( (U1+0.5)*(A2+0.5) ) ;
	  else
	    odds = (double)( A1*U2 ) / (double)( U1*A2 ) ;

	  if ( ! par::fisher_test )
	    pvalue = chiprobP(original,1);
	}

#ifdef DEBUG
	{

	  // Now display results for this SNP

	  ASC << setw(4)  << locus[l]->chr << " "
	      << setw(par::pp_maxsnp) << locus[l]->name << " "
	      << setw(10) << locus[l]->bp << " "
	      << setw(4)  << locus[l]->allele1 << " ";
	  if ( A1+A2 != 0 )
	    ASC << setw(8)  << (double)A1/(double)(A1+A2) << " ";
	  else
	    ASC << setw(8)  << "NA" << " ";

	  if ( U1+U2 != 0 )
	    ASC << setw(8)  << (double)U1/(double)(U1+U2) << " ";
	  else
	    ASC << setw(8)  << "NA" << " ";

	  ASC << setw(4)  << locus[l]->allele2 << " " ;

	    {
	      if ( pvalue > -1 )
		ASC << setw(12) << original  << " "
		    << setw(12) << pvalue << " ";
	      else
		{
		  ASC << setw(12) << "NA"  << " "
		    << setw(12) << "NA" << " ";
		}
	    }

	  double zero=0;
	  if (odds != odds || odds == 1/zero || odds == -1/zero )
	    {
	      ASC << setw(12) << "NA" << " ";
	      if (par::display_ci)
		ASC << setw(12) << "NA" << " "
		    << setw(12) << "NA" << " ";
	    }
	  else
	    {
	      ASC << setw(12) << odds  << " " ;
	      if (par::display_ci)
		{
		  double lOR = log(odds);
		  double SE = sqrt(1/(double)A1 + 1/(double)A2 + 1/(double)U1 + 1/(double)U2);
		  double OR_lower = exp( lOR - par::ci_zt * SE );
		  double OR_upper = exp( lOR + par::ci_zt * SE );

		  ASC << setw(12) << OR_lower << " "
		      << setw(12) << OR_upper << " ";
		}
	    }
	  ASC << "\n";
	}
#endif

	//LogisticOutput T;
	//T.snp = locus[l];
	//T.p = pvalue;
	//T.odds = odds;
	//out.push_back(T);
	//if(odds != NA) odds = exp(odds);
	//cout << "snp " << locus[l]->name << " P=" << pvalue << " OR=" << odds << " " << 1/odds << "\n";

	if ( pvalue <= -1 ) { pvalue = NA; odds = NA; }

	double zero=0;
	if (odds != odds || odds == 1/zero || odds == -1/zero ) { pvalue = NA; odds = NA; }

	out.push_back(LogisticOutput(locus[l],locus[l]->allele1,odds,pvalue));


    skip_p1:

      // Next SNP
      s++;
      l++;

    }
}





