/*
 * File:   runPlink.cpp
 * Author: Apostolos Dimitromanolakis
 *
 * Created on October 17, 2008, 11:03 PM
 */

#ifndef _RUNPLINK_H
#define	_RUNPLINK_H

#include <iostream>
#include <fstream>
#include <string>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "utils.h"
#include "br2.h"

extern char * line;
extern int line_pos, curline;

using namespace std;

/**
 *  Class plinkModule: Integration with plink
 *
 *  Covers:
 *     Running plink
 *     Reading plink output files
 *     Reading text PED/MAP files
 *     Writing BED/BIM/FAM files
 *     Reading writing covariate files
 *
 */

class plinkModule
{
public:

    // default mode is to write binary files in SNP major mode
    const static int out_SNP_major = 1;

    string plinkLocation; // location of plink executable
    string plinkOptions; // extra options to pass to plink


    plinkModule()
    {
	init();
    }

    ~plinkModule();

    void init();

    double checkPlinkVersion(string scratch);

    void runPLink(string & args);

    void model(const string& binfile, const string& outfile, vector<ModelOutput> &out);

    void logistic(const string& binfile, vector<LogisticOutput> &results,
	    vector<Locus*> &locus);

    void linear(const string & binfile,
	    vector<LogisticOutput> &results, vector<Locus*> &locus);

    void logisticX(const string& binfile, const string& outfile, vector<LogisticOutput> &results,
	    vector<Locus*> &locus);

    void genotypic(const string& binfile, const string& outfile, vector<GenotypicOutput> &results,
	    vector<Locus*> &locus);

    void QTLAssoc(const string& binfile, vector<ModelOutput> &out, vector<Locus*> &locus);

    void runRplugin(const string & binfile,
    	vector<LogisticOutput> &results, vector<Locus*> &locus, string pluginFile);

    void readMapFile(const string & mapfile);
    void readPedFile(const string & pedfile);
    void readTextFiles(const string & pedfile, const string & mapfile);

    void writeBimFile(const string &file, vector<Locus*> & locus);
    void writeBedFile(const string &head, vector<Individual*> &persons, vector<Locus*> &locus);
    void writeBinFiles(const string &head, vector<Individual*> &persons, vector<Locus*> &locus,
	    int newfids);

    void printPersons();

    void addPlinkOption(const string &h);

    void readCovariateFile(string in);
    void writeCovariateFile(string outFileName, const vector<Individual*> &persons, int newfids);

    void readAltPhenotype(string fileName);
    void deleteMissingPersons();


private:
    double checkPhenotype(string pheno);

    // Helper function to skip whitespace and get next phenotype

    inline
    int nextGenoType()
    {
	char c;

	do {
	    c = line[line_pos++];
	    /*getc(in);*/
	}
	while (c == ' ' || c == '\t');

	if (c == '\n') {
	    curline++;
	    return -1;
	}
	return c;
    }

};

#endif	/* _RUNPLINK_H */

