/*
 * File:   utils.cpp
 * Author: Apostolos Dimitromanolakis
 *
 * Created on October 17, 2008, 11:03 PM
 *
 * Miscellaneous functions
 *
 *
 */

#include "utils.h"
#include <fstream>
#include <time.h>

/*
 * Print vector v
 */

void print_v(vector<double> &v)
{
    double n = v.size(), a = 0, x, c = 0;

    char s[100];
    for (int i = 0; i < n; i++) {
	x = v[i];
	if (x != NA) sprintf(s, "%.5g ", v[i]);
	else sprintf(s, "NA ");
	cout << s;
    }
    cout << "\n";
}

/*
 * Output vector v histogram
 */

void histogram(vector<double> &v, int nbins, string file)
{

    double n = v.size(), c = 0;
    double max = v[0], min = v[0];

    vector<int> bins;
    bins.resize(nbins);

    for (int i = 0; i < n; i++) {
	if (v[i] > max) max = v[i];
	if (v[i] < min) min = v[i];
    }

    for (int i = 0; i < nbins; i++) bins[i] = 0;

    for (int i = 0; i < n; i++) {
	int b = (int) (0.0 + nbins * (v[i] - min) / max);
	bins[b]++;
    }

    ofstream f;
    f.open(file.c_str());

    for (int i = 0; i < nbins; i++) {
	f << (min + ((i - 0.0) * max / nbins)) << " " << bins[i] << "\n";
    }

}

/*
 * Compute average of vector v
 */

double avg(vector<double> &v)
{
    double a = 0, x, count = 0;
    int n = v.size();

    if (n == 0) {
	cout << "Empty vector for average!\n";
	return NA;
    }

    for (int i = 0; i < n; i++) {
	x = v[i];
	if (x != NA) {
	    a += x;
	    count++;
	}
    }
    return a / count;
}

/*
 * Compute Standard Deviation of vector v
 */

double sd(vector<double> &v)
{
    double vsize = v.size();
    double sum = 0, sumsq = 0, val;
    int n = 0;

    for (int i = 0; i < vsize; i++) {
	val = v[i];
	if (val == NA) continue;

	sum += val;
	sumsq += val * val;
	n++;
    }

    if (n < 2) return NA;

    double x_bar = sum / n;
    double x_bar_sq = x_bar * x_bar;

    double var = (sumsq / n - x_bar_sq) * n / (n - 1);
    return sqrt(var);
}


// Versatile function to print a table with every column aligned.
// Finds max size of each column and then prints needed white space.

/* data[i] is the whole column i
 max size[i] = max_j (data[i][j])

 */
void printTable(ostream &cout, const vector<vector<string> > &data)
{
    vector<int> col; // Column lengths

    string sep = "  ";

    int i, j;

    // Find max sizes for each column
    // max size[i] = max_i (data[j][i])

    for (i = 0; i < data[0].size(); i++) {
	int max = -999;
	for (j = 0; j < data.size(); j++)
	    max = std::max(max, (int) data[j][i].length());

	col.push_back(max);
    }

    char str[1024], fmt[128];
    // We assume number of entries is the size of vector at column 0

    for (j = 0; j < data.size(); j++) {

	for (i = 0; i < data[j].size(); i++) {

	    // Workaround to set first column left aligned
	    if (i != 0) sprintf(fmt, "%%%ds", (int) col[i]);
	    else sprintf(fmt, "%%-%ds", (int) col[i]);

	    sprintf(str, fmt, data[j][i].c_str());

	    cout << str << sep;
	}
	cout << "\n";
    }
}

/*
 * Timing functions
 *
 */

/*
 #include <sys/time.h>
 #include <time.h>

 double time0, utime0;


 void initTime() {
 struct timeval tv;
 gettimeofday(&tv, NULL);
 float t = tv.tv_sec + (double) tv.tv_usec / 1000000.0;
 time0 = tv.tv_sec;
 utime0 = tv.tv_usec;

 }

 double getElapsed() {
 struct timeval tv;
 gettimeofday(&tv, NULL);
 float t = (tv.tv_sec - time0);

 t += (double) (tv.tv_usec - utime0) / 1000000.0;
 return t;
 }

 void printTime(string s) {
 struct timeval tv;
 gettimeofday(&tv, NULL);
 float t = (tv.tv_sec - time0);

 t += (double) (tv.tv_usec - utime0) / 1000000.0;

 char out[100];
 sprintf(out, "%.4f", t);
 //LOG << s << " time=" << out << " ";
 initTime();
 }
 */

/*
 * Not used right now
 *
 */

/*int ind[8000], tot[8000];

 void testBScounts(int maxj)
 {
 srand(time(NULL));
 int j = 0;

 int gmax = 0;

 while (j <= maxj)
 {
 int i;
 for (i = 0; i < 5000; i++)
 ind[i] = 0;

 for (i = 0; i < 5000; i++)
 ind[rand() % 5000]++;

 //int k =0;
 // for (i = 0; i < 5000; i++) if(ind[i] > 0) k++;
 //cout << "=>" << ((double)k/5000) << "\n";

 for (i = 0; i < 5000; i++)
 if (tot[i] < ind[i])
 tot[i] = ind[i];
 j++;

 }

 int i;
 int max = 0, sum = 0;
 for (i = 0; i < 5000; i++)
 {
 sum += tot[i];
 if (tot[i] > max)
 max = tot[i];
 }

 //for(i=0;i<5000;i++) cout << ind[i] << " ";

 if (max > gmax)
 gmax = max;

 cout << gmax << " ";
 cout << sum << " vs " << ((j - 1) * 5000) << " " << (1 / ((double) sum
 / ((j - 1) * 5000))) << "\n";
 j = 1000;

 }
 */

