/*
 * File:   utils.h
 * Author: Apostolos Dimitromanolakis
 *
 * Created on October 19, 2008, 12:07 AM
 */

#ifndef _UTILS_H
#define	_UTILS_H

// This is needed by visual c++ to avoid annoying warnings
#define _CRT_SECURE_NO_WARNINGS 1

#include <iostream>
#include <string>
#include <vector>

#include <iomanip>
#include <deque>
#include <sstream>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

using namespace std;

// Double value used to signify NA (data not available).
const double NA = 10e80;

#ifndef M_E
#define M_E		2.7182818284590452354	/* e */
#endif

namespace br2options {
    extern string tmpdir;
    extern int debug;
};

using namespace br2options;

typedef unsigned int uint;


// Computes average and standard deviation of a set of doubles with little memory overhead
class statistics
{
    int n;
    double sum, sumsq;

    statistics() {
	n = 0;
	sum = 0;
	sumsq = 0;
    }

    void add(double a) {
	if (a == NA) return;
	sum += a;
	sumsq += a*a;
	++n;
    }

    double average() { // return average of samples
	if(n == 0) return NA;
	return sum / n;
    }

    double sd() { // return standard deviation
	if (n < 2) return NA;
	double x_bar = sum / n;
	double x_bar_sq = x_bar * x_bar;

	double var = (sumsq / n - x_bar_sq) * n / (n - 1);
	return sqrt(var);
    }
};

inline string toString(int a)
{
    char v[50];
    sprintf(v, "%d", a);
    string t(v);
    return t;
}

inline string toString(double a, const string & fmt)
{
    char v[100];

    if (a == NA)
	sprintf(v, "NA"); // BUG: Actually should pad here according to fmt
    else
	sprintf(v, fmt.c_str(), a);

    string t(v);
    return t;
}

inline string toString(double a)
{
    return toString(a, "%g");
}

inline double toDouble(char s[])
{
    if (strcmp(s, "NA") == 0) return NA;
    else return atof(s);
}

inline double min(double a, double b)
{
    if (a == NA) return b;
    if (b == NA) return NA;

    if (a < b) return a;
    else return b;
}

inline double max(vector<double> &v)
{
    if(v.size() == 0) return NA;

    double max = v[0];
    for (uint i = 0; i < v.size(); i++)
	if (v[i] > max && v[i] != NA) max = v[i];
    return max;
}

inline double min(vector<double> &v)
{
    if(v.size() == 0) return NA;

    double min = v[0];
    for (uint i = 0; i < v.size(); i++)
	if (v[i] < min && v[i] != NA) min = v[i];
    return min;
}

/* Delete a file using a string argument */
inline int remove(string s)
{
    return remove(s.c_str());
}

/* Split a string into tokens */
inline vector<string> split(string s)
{
    vector<string> list;
    istringstream a(s);
    string t;

    while(a >> t) { list.push_back(t); }

    return list;
}

/* The following functions are defined in utils.cpp */

void gracefulExit(int retval);
void cleanup();
double avg(vector<double> &v);
double median(vector<double> &v);
double sd(vector<double> &v);
void print_v(vector<double> &v);
void histogram(vector<double> &v, int nbins, string file);
void printTable(ostream &out, const vector<vector<string> > &data);

#endif	/* _UTILS_H */

